from fluoriclogppka.ml_part.data_preparation.smiles_to_features import Featurizer
from fluoriclogppka.ml_part.inference.inference import Inference

from fluoriclogppka.ml_part.services.features import __all__
from fluoriclogppka.ml_part.constants import Target, Identificator



from rdkit import Chem
from rdkit.Chem import Descriptors

from fluoriclogppka.ml_part.constants import Identificator
from fluoriclogppka.ml_part.constants import ALL_SUBMOLS

def amount_of_hydrogen_in_neighbors(mol, atom_idx):
    """
    Calculate the number of hydrogen atoms in the neighbors of a given atom.

    Args:
        mol (rdchem.Mol): RDKit molecule.
        atom_idx (int): Index of the target atom.

    Returns:
        int: Number of hydrogen atoms in the neighboring atoms.
    """
    atom = mol.GetAtomWithIdx(atom_idx)

    neighbors = atom.GetNeighbors()
    
    amount_of_hydrogen = 0
    
    for neighbor in neighbors:
        if neighbor.GetSymbol() == 'H':
            amount_of_hydrogen += 1

    return amount_of_hydrogen

def calculate_identificator(mol) -> Identificator:
    """
    Identify the type of functional group present in the molecule - COOH, NH or NH2.

    Returns:
        Identificator: Type of functional group.

    Raises:
        TypeError: If the molecule doesn't match any expected functional group.
    """
    carboxile_submol = Chem.MolFromSmiles('CC=O')
    nitro_amine_submol = Chem.MolFromSmiles('CN')

    carboxile_matches = mol.GetSubstructMatches(carboxile_submol)
    nitro_amine_matches = mol.GetSubstructMatches(nitro_amine_submol)

    if len(carboxile_matches) > 0:
        return Identificator.carboxilic_acid
    
    n_atom_index = nitro_amine_matches[0][1]
    amount_of_hydrogen = amount_of_hydrogen_in_neighbors(mol, n_atom_index)

    if amount_of_hydrogen == 2:
        return Identificator.primary_amine
    elif amount_of_hydrogen == 1:
        return Identificator.secondary_amine
    
    raise TypeError("Wrong molecule")

def calculate_linear_path_f_to_fg(smiles):
    """
    Calculate amount of linear paths from fluorine (F) to the functional group (FG) based on the provided SMILES.

    Args:
        smiles (str): SMILES representation of the molecule.

    Returns:
        int: The total number of matches found between substructures and the molecule.
    """    
    mol = Chem.MolFromSmiles(smiles, sanitize=True) 
    mol = Chem.AddHs(mol) 
    
    total_matches = 0
    for _, sub_SMILES_array in ALL_SUBMOLS.items():
        for sub_SMILES in sub_SMILES_array:
            
            submol = Chem.MolFromSmiles(sub_SMILES, sanitize=False)
            matches = mol.GetSubstructMatches(submol)

            if len(matches) > 0:
                total_matches += len(matches)
    
    return total_matches

def calculate_molecular_weight(SMILES):
    """
    Calculate the molecular weight of the molecule based on the provided SMILES.

    Args:
        SMILES (str): SMILES representation of the molecule.

    Returns:
        float: The calculated molecular weight.
    """
    mol = Chem.MolFromSmiles(SMILES)
    mol = Chem.AddHs(mol)
    
    molecular_weight = Descriptors.MolWt(mol)
    
    return molecular_weight



from rdkit import Chem
from rdkit.Chem import AllChem


class Molecule2DFeaturesService:
    """
    Class that represents a 2D molecule features.

    This class provides functionality to obtain 2d molecules features. 

    Attributes:
        mol: Rdkit molecule with single conformer.
        mol_num_cycles (int): Amount of cycles in the molecule.
        atoms_num_in_cycles (float): Ratio of amount of atoms in rings to rings amount.
        chirality (int): Amount of chiral centers in the molecule.
        features_2d_dict (dict(str, float)): Feature name to feature value.

    Methods:
        prepare_molecule(): Creates rdkit molecule with single conformers from smiles.
        mol_cycles_amount(): Get amount of rings in the molecule.
        atoms_num_in_cycles_divide_by_amount_cycles(): Get ratio of amount of atoms in rings to rings amount.
        get_amount_of_chiral_centers(): Amount of chiral centers in molecule.
    """
    def __init__(self,
                 SMILES):
        """
        Initialize the Molecule2DFeaturesService instance and calculates molecule features,
        such as: amount of cycles, ratio of amount of atoms in rings to rings amount, amount of chiral centers.

        Args:
            SMILES (str): String representation of a molecule.
        """
        self.mol = Molecule2DFeaturesService.prepare_molecule(SMILES)

        self.mol_num_cycles = self.mol_cycles_amount()
        self.atoms_num_in_cycles = self.atoms_num_in_cycles_divide_by_amount_cycles()
        self.chirality = self.get_amount_of_chiral_centers()

        self.features_2d_dict = {
            "mol_num_cycles": self.mol_num_cycles,
            "avg_atoms_in_cycle": self.atoms_num_in_cycles,
            "chirality": self.chirality,
        }


    @staticmethod
    def prepare_molecule(SMILES):
        """
        Create rdkit 3d molecule from SMILES with conformer.
        
        Args:
            SMILES (str): String representation of a molecule. 
            
        Returns:
            mol: Rdkit 3D molecule with single conformer.
        """
        mol = Chem.MolFromSmiles(SMILES)
        mol = Chem.AddHs(mol)

        AllChem.EmbedMolecule(mol, randomSeed=42)

        return mol
    

    def mol_cycles_amount(self):
        """
        Calculate amount of cycles in the molecule.
            
        Returns:
            num_rings (int): amount of rings.
        """
        sssr = Chem.GetSSSR(self.mol)

        num_rings = len(sssr)
        return num_rings

    
    def atoms_num_in_cycles_divide_by_amount_cycles(self):
        """
        Calculate ratio amount of atoms in rings to rings amount.
            
        Returns:
            (float): atoms_num_in_cycles / amount_cycles.
        """
        sssr = Chem.GetSSSR(self.mol)
        amount_cycles = len(sssr)
        
        if amount_cycles == 0:
            return 0

        atoms_idxs = set()

        for i, ring in enumerate(sssr):
            for atom_index in ring:
                atoms_idxs.add(atom_index)
        
        atoms_num_in_cycles = len(atoms_idxs)
        
        return atoms_num_in_cycles / amount_cycles
    

    def get_amount_of_chiral_centers(self):
        """
        Calculate amount of chiral centers in molecule.
            
        Returns:
            amount_of_chiral_centers (int): amount of chiral centers in the molecule.
        """
        mol = Chem.AddHs(self.mol)
        AllChem.EmbedMolecule(mol)
        
        Chem.AssignAtomChiralTagsFromStructure(mol)
        chirality_centers = Chem.FindMolChiralCenters(mol)
        amount_of_chiral_centers = len(chirality_centers)
        
        return amount_of_chiral_centers



from setuptools import setup, find_packages

setup(
  name = 'fluoriclogppka',
  packages = find_packages(), 
  include_package_data = True,

  version = '0.1.1',
  license = 'MIT',
  description = 'Tool for pKa, logP prediction',
  author = 'Blackthorn.ai',
  url = 'https://github.com/blackthorn-ai/fluoricLogPpKa',
  download_url = 'https://github.com/blackthorn-ai/fluoricLogPpKa/dist/fluoriclogppka-0.1.1.tar.gz',
  keywords = ['pKa', 'logP', 'tool'],
  
  install_requires=[ 
          'certifi==2024.2.2',
          'charset-normalizer==3.3.2',
          'future==1.0.0',
          'h2o==3.44.0.3',
          'idna==3.6',
          'mordred==1.2.0',
          'networkx==2.8.8',
          'numpy==1.26.4',
          'pandas==2.2.0',
          'pillow==10.2.0',
          'pyarrow==15.0.0',
          'python-dateutil==2.8.2',
          'pytz==2024.1',
          'rdkit==2023.9.5',
          'requests==2.31.0',
          'six==1.16.0',
          'tabulate==0.9.0',
          'tzdata==2024.1',
          'urllib3==2.2.1'
      ],
  classifiers=[
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3.11',
  ],
)



import pandas as pd

from fluoriclogppka.ml_part.exceptions import FeatureNotFoundError

from fluoriclogppka.ml_part.constants import Target
from fluoriclogppka.ml_part.constants import LOGP_FEATURES, PKA_FEATURES
from fluoriclogppka.ml_part.constants import CONVERT_FEATURE_TO

from fluoriclogppka.ml_part.services.molecule_3d_features_service import Molecule3DFeaturesService
from fluoriclogppka.ml_part.services.molecule_2d_features_service import Molecule2DFeaturesService
from fluoriclogppka.ml_part.services.mordred_features_service import MordredFeaturesService

class Featurizer:
    """
    A class for preparing data for chemical property prediction.

    This class prepares data from Enamine dataset for predicting chemical properties such as pKa or logP.

    Attributes:
        SMILES (str): The SMILES string representing the molecule.
        target_value (Target): The target property to predict (pKa or logP).
        row_from_enamine_dataset (pd.Series): A row from the Enamine Dataframe containing molecule information.

    Methods:
        extract_all_features(): Extracts all posible features for the molecule from 
            rdkit, mordred and our dataset.
        extract_required_features(): Extracts only the required features for predicting pKa or LogP.
        prepare_features_for_model(): Converts string molecule features to int. 
    """
    def __init__(self, 
                 SMILES: str,
                 target_value: Target,
                 ) -> None:
        """
        Initialize the PrepareFluorineData object.

        Args:
            SMILES (str): The SMILES string representing the molecule.
            target_value (Target): The target property to predict (pKa or logP).
            row_from_enamine_dataset (pd.Series): A row from the Enamine dataset containing molecule information.
        """
        self.SMILES = SMILES
        self.target_value = target_value

        if target_value == Target.pKa:
            self.required_features = PKA_FEATURES
        elif target_value == Target.logP:
            self.required_features = LOGP_FEATURES

        self.all_features_dict = self.extract_all_features()

        self.required_features_for_predict = self.extract_required_features()

        self.features_for_predict = self.prepare_features_for_model()

    def extract_all_features(self):
        """
        Extracts all posible features for the molecule from rdkit, mordred and our dataset.

        Returns:
            dict: A dictionary containing all extracted features from rdkit, mordred and Enamine dataset.
        """
        all_features = {}

        mordredFeaturesService = MordredFeaturesService(self.SMILES)
        all_features.update(mordredFeaturesService.mordred_features_dict)

        moleculeFeatures2dService = Molecule2DFeaturesService(self.SMILES)
        all_features.update(moleculeFeatures2dService.features_2d_dict)

        moleculeFeatures3dService = Molecule3DFeaturesService(smiles=self.SMILES,
                                                              target_value=self.target_value)
        all_features.update(moleculeFeatures3dService.features_3d_dict)

        return all_features

    def extract_required_features(self):
        """
        Extracts only the required features for predicting target value(pKa or LogP).

        Returns:
            dict: A dictionary containing the required features and their values for prediction.
        """
        for feature_name in self.required_features:
            if feature_name not in self.all_features_dict:
                raise FeatureNotFoundError(feature_name)
            
        required_features_dict = {feature: self.all_features_dict[feature] for feature in self.required_features} 
        return required_features_dict
    
    def prepare_features_for_model(self):
        """
        Prepare features for model prediction.

        This method prepares features for model prediction based on the required features
        for prediction and their corresponding int values. It applies any necessary conversions
        to the feature values before returning the prepared features.

        Returns:
            dict: A dictionary containing the prepared features for model prediction.
        """
        features_for_predict = self.required_features_for_predict.copy()
        
        for feature_name, value in features_for_predict.items():
            if feature_name in CONVERT_FEATURE_TO:
                features_for_predict[feature_name] = CONVERT_FEATURE_TO[feature_name][value]

        return features_for_predict
